@@ -337,18 +337,23 @@ meta_monitor_manager_xrandr_read_current (MetaMonitorManager *manager)
       dpms_enabled)
     {
       switch (dpms_state)
-	{
-	case DPMSModeOn:
-	  manager->power_save_mode = META_POWER_SAVE_ON;
-	case DPMSModeStandby:
-	  manager->power_save_mode = META_POWER_SAVE_STANDBY;
-	case DPMSModeSuspend:
-	  manager->power_save_mode = META_POWER_SAVE_SUSPEND;
-	case DPMSModeOff:
-	  manager->power_save_mode = META_POWER_SAVE_OFF;
-	default:
-	  manager->power_save_mode = META_POWER_SAVE_UNKNOWN;
-	}
+        {
+        case DPMSModeOn:
+          manager->power_save_mode = META_POWER_SAVE_ON;
+          break;
+        case DPMSModeStandby:
+          manager->power_save_mode = META_POWER_SAVE_STANDBY;
+          break;
+        case DPMSModeSuspend:
+          manager->power_save_mode = META_POWER_SAVE_SUSPEND;
+          break;
+        case DPMSModeOff:
+          manager->power_save_mode = META_POWER_SAVE_OFF;
+          break;
+        default:
+          manager->power_save_mode = META_POWER_SAVE_UNKNOWN;
+          break;
+        }
     }
   else
     {
@@ -527,7 +532,7 @@ meta_monitor_manager_xrandr_read_current (MetaMonitorManager *manager)
 	  */
 	  for (j = 0; j < (unsigned)output->nclone; j++)
 	    {
-	      meta_output->possible_clones = GINT_TO_POINTER (output->clones[j]);
+	      meta_output->possible_clones[j] = GINT_TO_POINTER (output->clones[j]);
 	    }
 
 	  meta_output->is_primary = ((XID)meta_output->output_id == primary_output);
@@ -772,17 +777,51 @@ meta_monitor_manager_xrandr_apply_configuration (MetaMonitorManager *manager,
         {
           MetaMonitorMode *mode;
           XID *outputs;
-          int j, n_outputs;
+          unsigned int j, n_outputs;
           int width, height;
           Status ok;
+          unsigned long old_controlled_mask;
+          unsigned long new_controlled_mask;
 
           mode = crtc_info->mode;
 
           n_outputs = crtc_info->outputs->len;
           outputs = g_new (XID, n_outputs);
 
+          old_controlled_mask = 0;
+          for (j = 0; j < manager->n_outputs; j++)
+            {
+              MetaOutput *output;
+
+              output = &manager->outputs[j];
+
+              if (output->crtc == crtc)
+                old_controlled_mask |= 1UL << j;
+            }
+
+          new_controlled_mask = 0;
           for (j = 0; j < n_outputs; j++)
-            outputs[j] = ((MetaOutput**)crtc_info->outputs->pdata)[j]->output_id;
+            {
+              MetaOutput *output;
+
+              output = ((MetaOutput**)crtc_info->outputs->pdata)[j];
+
+              output->is_dirty = TRUE;
+              output->crtc = crtc;
+              new_controlled_mask |= 1UL << j;
+
+              outputs[j] = output->output_id;
+            }
+
+          if (crtc->current_mode == mode &&
+              crtc->rect.x == crtc_info->x &&
+              crtc->rect.y == crtc_info->y &&
+              crtc->transform == crtc_info->transform &&
+              old_controlled_mask == new_controlled_mask)
+            {
+              /* No change */
+              goto next;
+            }
 
           meta_error_trap_push (meta_get_display ());
           ok = XRRSetCrtcConfig (manager_xrandr->xdisplay,
@@ -801,11 +840,9 @@ meta_monitor_manager_xrandr_apply_configuration (MetaMonitorManager *manager,
                             (unsigned)(crtc->crtc_id), (unsigned)(mode->mode_id),
                             mode->width, mode->height, (float)mode->refresh_rate,
                             crtc_info->x, crtc_info->y, crtc_info->transform);
-              continue;
+              goto next;
             }
 
-          g_free (outputs);
-
           if (meta_monitor_transform_is_rotated (crtc_info->transform))
             {
               width = mode->height;
@@ -824,15 +861,8 @@ meta_monitor_manager_xrandr_apply_configuration (MetaMonitorManager *manager,
           crtc->current_mode = mode;
           crtc->transform = crtc_info->transform;
 
-          for (j = 0; j < n_outputs; j++)
-            {
-              MetaOutput *output;
-
-              output = ((MetaOutput**)crtc_info->outputs->pdata)[j];
-
-              output->is_dirty = TRUE;
-              output->crtc = crtc;
-            }
+        next:
+          g_free (outputs);
         }
     }
 
@@ -944,11 +974,57 @@ meta_monitor_manager_xrandr_handle_xevent (MetaMonitorManager *manager,
 					   XEvent             *event)
 {
   MetaMonitorManagerXrandr *manager_xrandr = META_MONITOR_MANAGER_XRANDR (manager);
+  MetaOutput *old_outputs;
+  MetaCRTC *old_crtcs;
+  MetaMonitorMode *old_modes;
+  int n_old_outputs;
 
   if ((event->type - manager_xrandr->rr_event_base) != RRScreenChangeNotify)
     return FALSE;
 
   XRRUpdateConfiguration (event);
+
+  /* Save the old structures, so they stay valid during the update */
+  old_outputs = manager->outputs;
+  n_old_outputs = manager->n_outputs;
+  old_modes = manager->modes;
+  old_crtcs = manager->crtcs;
+
+  manager->serial++;
+  meta_monitor_manager_xrandr_read_current (manager);
+
+  /* Check if the current intended configuration has the same outputs
+     as the new real one, or if the event is a result of an XRandR call.
+     If so, we can go straight to rebuild the logical config and tell
+     the outside world.
+     Otherwise, this event was caused by hotplug, so give a chance to
+     MetaMonitorConfig.
+
+     Note that we need to check both the timestamps and the list of
+     outputs, because the X server might emit spurious events with
+     new configTimestamps (bug 702804), and the driver may have
+     changed the EDID for some other reason (old broken qxl and vbox
+     drivers...).
+  */
+  if (manager_xrandr->resources->timestamp >= manager_xrandr->resources->configTimestamp ||
+      meta_monitor_config_match_current (manager->config, manager))
+    {
+      /* This will be a no-op if the change was from our side, as
+         we already called it in the DBus method handler */
+      meta_monitor_config_update_current (manager->config, manager);
+
+      meta_monitor_manager_rebuild_derived (manager);
+    }
+  else
+    {
+      if (!meta_monitor_config_apply_stored (manager->config, manager))
+        meta_monitor_config_make_default (manager->config, manager);
+    }
+
+  meta_monitor_manager_free_output_array (old_outputs, n_old_outputs);
+  g_free (old_modes);
+  g_free (old_crtcs);
+
   return TRUE;
 }
 
