@@ -54,9 +54,6 @@ item(WM_WINDOW_ROLE)
 item(UTF8_STRING)
 item(WM_ICON_SIZE)
 item(_KWM_WIN_ICON)
-item(_MUTTER_RELOAD_THEME_MESSAGE)
-item(_MUTTER_SET_KEYBINDINGS_MESSAGE)
-item(_MUTTER_TOGGLE_VERBOSE)
 item(_MUTTER_HINTS)
 item(_GTK_THEME_VARIANT)
 item(_GTK_HIDE_TITLEBAR_WHEN_MAXIMIZED)
@@ -66,6 +63,7 @@ item(_GTK_APPLICATION_OBJECT_PATH)
 item(_GTK_WINDOW_OBJECT_PATH)
 item(_GTK_APP_MENU_OBJECT_PATH)
 item(_GTK_MENUBAR_OBJECT_PATH)
+item(_GTK_FRAME_EXTENTS)
 item(_GNOME_WM_KEYBINDINGS)
 item(_GNOME_PANEL_ACTION)
 item(_GNOME_PANEL_ACTION_MAIN_MENU)
@@ -74,6 +72,7 @@ item(_MUTTER_TIMESTAMP_PING)
 item(_MUTTER_FOCUS_SET)
 item(_MUTTER_SENTINEL)
 item(_MUTTER_VERSION)
+item(_MUTTER_PRESENTATION_OUTPUT)
 item(WM_CLIENT_MACHINE)
 item(MANAGER)
 item(TARGETS)
@@ -81,6 +80,7 @@ item(MULTIPLE)
 item(TIMESTAMP)
 item(VERSION)
 item(ATOM_PAIR)
+item(BACKLIGHT)
 
 /* Oddities: These are used, and we need atoms for them,
  * but when we need all _NET_WM hints (i.e. when we're making
